from collections import OrderedDict

all_routines = OrderedDict([('pflow', ['PFlow']),
                            ('tds', ['TDS']),
                            ('eig', ['EIG']),
                            ])
